import os.path
from pathlib import Path
import json
import wave
import struct

# normal_files = ["hello.wav", "my.wav", "name.wav", "is.wav", "vox.wav"
#                 "_comma.wav", "_period.wav"]
# empty_files = []
# no_punct_files = ["hello.wav", "my.wav", "name.wav", "is.wav", "vox.wav"]
# inconst_format_files = ["hello.mp3", "my.wav", "name", "is.wav", "vox.mp4",
#                         "_comma.wav", "_period.wav"]
# test_files = {"normal": normal_files, "empty": [],
#               "no_punct": no_punct_files, "inconst_format": inconst_format_files,
#               }


def create_test_files(audio_path, files_dict):
    audio_path_p = Path(audio_path)
    audio_path_p.mkdir(parents=True, exist_ok=True)

    for path_name in files_dict.keys():
        path_name_d = audio_path_p.joinpath(path_name)
        path_name_d.mkdir(parents=True, exist_ok=True)

        for file in files_dict[path_name]:
            filepath = path_name_d.joinpath(file)
            # savefile = open(filepath, "w")
            savefile = wave.open(str(filepath), "w")

            # channels, datasize (16 bit), sample rate, number of samples
            savefile.setparams((1, 2, 11025, 500, "NONE", "Uncompressed"))
            savefile.writeframes(struct.pack('h', 1))
            savefile.close()


def clean_test_files(audio_path, export_path, db_exports_only=False):
    audio_path_p = Path(audio_path)
    export_path_p = Path(export_path)

    if audio_path_p.is_dir():
        test_dirs = list(x for x in audio_path_p.iterdir() if x.is_dir())
        for dirc in test_dirs:

            # Delete info if it exists
            infopath_dir = dirc.joinpath("info/")
            if infopath_dir.is_dir() and not db_exports_only:
                infopath_file = infopath_dir.joinpath("info.json")
                os.remove(infopath_file)
                os.rmdir(infopath_dir)

            # Delete db if it exists
            db_dir = dirc.joinpath("db/")
            if db_dir.is_dir():
                db_file = db_dir.joinpath("db.json")
                os.remove(db_file)
                os.rmdir(db_dir)

            if not db_exports_only:
                files = list(dirc.iterdir())

                for file in files:
                    os.remove(file)
                os.rmdir(dirc)
        if not db_exports_only:
            os.rmdir(audio_path)

    if export_path_p.is_dir():
        voice_export_dirs = list(export_path_p.iterdir())
        for voice_export_dir in voice_export_dirs:
            files = list(voice_export_dir.iterdir())
            for file in files:
                os.remove(file)
            os.rmdir(voice_export_dir)
        os.rmdir(export_path_p)


def create_info(info_dict, audio_path):
    info_dir = Path(audio_path).joinpath("info/")

    info_dir.mkdir(parents=True, exist_ok=True)

    info_file = info_dir.joinpath("info.json")

    with open(info_file, 'w') as file:
        json.dump(info_dict, file)


# def create_sentence(export_path, sentence_filenames):
#     export_path_p = Path(export_path)

#     export_path_p.mkdir(parents=True, exist_ok=True)

#     for sentence in sentence_filenames:
#         filepath = export_path_p.joinpath(sentence)
#         savefile = open(filepath, "w")
#         savefile.close()


# if __name__ == "__main__":
#     create_sentence("./tests/tempexport", "hello-world.wav")



import os
from pathlib import Path
from hlvox.voice import Voice


class Manager:
    def __init__(self, voices_path, exports_path):
        self.voices_path = Path(voices_path)
        self.exports_path = Path(exports_path)
        self.voices = self._load_voices(self.voices_path)
        # self._create_export_dirs(self.voices.keys())

    def _load_voices(self, path):
        voices = {}
        voice_folders = list(x for x in path.iterdir() if x.is_dir())
        for voice_folder in voice_folders:
            export_path = self.exports_path / voice_folder.name
            export_path.mkdir(parents=True, exist_ok=True)
            new_voice = Voice(path=voice_folder, export_path=export_path)
            voices[new_voice.name] = new_voice
        return voices

    # def _create_export_dirs(self, voices):
    #     for voice in voices:
    #         path = self.exports_path / voice
    #         if not path.exists():
    #             os.mkdir(path)

    def get_voice_names(self):
        """Gets names of available voices

        Returns:
            list -- list of voice name strings
        """

        return list(self.voices.keys())

    def get_voice(self, name):
        """Get voice of requested name

        Args:
            name ({string}): name of voice to get

        Returns:
            {voxvoice}: requested voice
        """
        if name in self.voices:
            return self.voices[name]
        else:
            return None

    def exit(self):
        for voice_name in self.voices:
            self.voices[voice_name].exit()



import utils as th
import unittest
from pathlib import Path
from hlvox import Voice, Manager

test_filepath = "./tests/tempfiles"
test_filepath_p = Path(test_filepath)
test_exportpath = "./tests/tempexport"
test_exportpath_p = Path(test_exportpath)

test_files = {
    "v1": ["hello.wav"],
    "v2": ["hello.wav"],
    "v3": ["hello.wav"]
}


class test_voice_import(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        th.clean_test_files(test_filepath, test_exportpath)

        th.create_test_files(test_filepath, test_files)

    @classmethod
    def tearDownClass(cls):
        th.clean_test_files(test_filepath, test_exportpath)

    def setUp(self):
        self.tu = Manager(test_filepath, test_exportpath)

    def tearDown(self):
        self.tu.exit()

    def test_export_dir_creation(self):
        exp_export_dirs = [test_exportpath_p.joinpath("v1"),
                           test_exportpath_p.joinpath("v2"),
                           test_exportpath_p.joinpath("v3")]

        export_dirs = test_exportpath_p.iterdir()

        self.assertCountEqual(exp_export_dirs, export_dirs)

    def test_voice_list(self):

        exp_voice_list = ["v1", "v2", "v3"]

        self.assertCountEqual(exp_voice_list, self.tu.get_voice_names())


class test_get_voice(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        th.clean_test_files(test_filepath, test_exportpath)

        th.create_test_files(test_filepath, test_files)

    @classmethod
    def tearDownClass(cls):
        th.clean_test_files(test_filepath, test_exportpath)

    def setUp(self):
        self.tu = Manager(test_filepath, test_exportpath)

    def tearDown(self):
        self.tu.exit()

    def test_get_voice(self):
        voice = self.tu.get_voice("v1")

        self.assertEqual("v1", voice.name)

    def test_get_wrong_voice(self):
        voice = self.tu.get_voice("nope")

        self.assertEqual(None, voice)



from pathlib import Path
import utils as th
from hlvox import Voice
import unittest


# Stand-in files for testing
normal_files = ["hello.wav", "my.wav", "name.wav", "is.wav", "vox.wav",
                "_comma.wav", "_period.wav"]
empty_files = []
no_punct_files = ["hello.wav", "my.wav", "name.wav", "is.wav", "vox.wav"]
inconst_format_files = ["hello.mp3", "my.wav", "name", "is.wav", "vox.mp4",
                        "_comma.wav", "_period.wav"]
alarm_files = ["hello.wav", "my.wav", "name.wav", "is.wav", "vox.wav",
               "_comma.wav", "_period.wav", "dadeda.wav", "woop.wav"]
alph_files = ["a.wav", "b.wav", "c.wav", "_comma.wav", "_period.wav"]
test_files = {"normal": normal_files, "empty": empty_files,
              "no_punct": no_punct_files, "inconst_format": inconst_format_files,
              "alarm": alarm_files, "alph": alph_files}


test_filepath = "./tests/tempfiles"
test_filepath_p = Path(test_filepath)
test_exportpath = "./tests/tempexport"
test_exportpath_p = Path(test_exportpath)

# Example info file
ex_info_file = {"alarms": ["dadeda", "woop"]}


test_files_paths = {name: test_filepath_p.joinpath(name)
                    for name in test_files.keys()}


class test_file_handling(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        th.clean_test_files(test_filepath, test_exportpath)

        th.create_test_files(test_filepath, test_files)

    def test_empty_files(self):
        with self.assertRaises(Exception) as error:
            Voice(test_files_paths["empty"],
                  test_exportpath_p.joinpath("empty"))

        self.assertEqual(str(error.exception), "No words found")

    # ?: Removed for now, could use some rethinking
    # def test_no_punct_files(self):
    #     # TODO: test no comma, no period, both

    #     with self.assertRaises(Exception) as error:
    #         Voice(test_files_paths["no_punct"], test_exportpath)

    #     self.assertEqual(str(error.exception), "Incomplete Punctuation")

    def test_inconsistent_format(self):

        with self.assertRaises(Exception) as error:
            Voice(test_files_paths["inconst_format"],
                  test_exportpath_p.joinpath("inconst_format"))

        self.assertEqual(str(error.exception), "Inconsistent Audio Formats")

    def test_audio_format(self):

        unit = Voice(test_files_paths["normal"],
                     test_exportpath_p.joinpath("normal"))

        self.assertEqual(unit.get_audio_format(), "wav")
        unit.exit()

    @classmethod
    def tearDownClass(cls):
        th.clean_test_files(test_filepath, test_exportpath)


class test_dict_contents(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        sub_test_files = ["normal", "alph"]
        th.create_test_files(test_filepath, dict(
            (key, test_files[key]) for key in sub_test_files))

    def setUp(self):
        self.tu = None

    def tearDown(self):
        self.tu.exit()

    def test_basic_dict(self):
        self.tu = Voice(test_files_paths["normal"],
                        test_exportpath_p.joinpath("normal"))

        expected_names = [
            name[:-4] for name in normal_files if name not in ["_comma.wav", "_period.wav"]]
        expected_names.extend([",", "."])

        tu_dict = self.tu.get_words()

        self.assertCountEqual(expected_names, tu_dict)

    def test_alphab(self):
        self.tu = Voice(test_files_paths["alph"],
                        test_exportpath_p.joinpath("alph"))

        expected_names = [",", ".", "a", "b", "c"]

        tu_dict = self.tu.get_words()

        self.assertEqual(expected_names, tu_dict)

    @classmethod
    def tearDownClass(cls):
        th.clean_test_files(test_filepath, test_exportpath)


class test_voice_info_loading(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        sub_test_files = ["alarm"]
        th.create_test_files(test_filepath, dict(
            (key, test_files[key]) for key in sub_test_files))
        th.create_info(ex_info_file, test_files_paths["alarm"])

    def setUp(self):
        self.tu = None

    def tearDown(self):
        self.tu.exit()

    def test_alarm_read(self):
        self.tu = Voice(test_files_paths["alarm"],
                        test_exportpath_p.joinpath("alarm"))
        print(self.tu.alarms)

        self.assertCountEqual(self.tu.alarms, ["dadeda", "woop"])

    @classmethod
    def tearDownClass(cls):
        th.clean_test_files(test_filepath, test_exportpath)


# TODO: sayable/unsayable might take the place of this


class test_sentence_lists(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        th.create_test_files(test_filepath, test_files)

    @classmethod
    def tearDownClass(cls):
        th.clean_test_files(test_filepath, test_exportpath)

    def setUp(self):
        self.tu = Voice(test_files_paths["normal"],
                        test_exportpath_p.joinpath("normal"))

    def tearDown(self):
        self.tu.exit()

    def test_empty_sentence(self):
        ret = self.tu.get_sentence_list("")

        self.assertEqual([], ret)

    def test_simple_sentence(self):
        ret = self.tu.get_sentence_list("hello")

        self.assertEqual(["hello"], ret)

    def test_simple_punct(self):
        ret = self.tu.get_sentence_list("hello, world")

        self.assertEqual(["hello", ","], ret)

    def test_comp_punct(self):
        ret = self.tu.get_sentence_list("hello , world. Vox , , says hi")

        self.assertEqual(["hello", ",", ".",
                          "vox", ",", ","], ret)

    def test_punct_only(self):
        ret = self.tu.get_sentence_list(",")

        self.assertEqual([","], ret)

    def test_no_space_punct(self):
        ret = self.tu.get_sentence_list(",.")

        self.assertEqual([",", "."], ret)

    def test_temp(self):
        ret = self.tu.get_sentence_list("hello. my name, is, vox")

        self.assertEqual(["hello", ".", "my", "name",
                          ",", "is", ",", "vox"], ret)

    def test_punct_location(self):
        # Not sure how to deal with types like ".hello"
        # for now it will treat it as just a period and throw out all the characters after it
        ret1 = self.tu.get_sentence_list("hello.")
        ret2 = self.tu.get_sentence_list(".hello")
        ret3 = self.tu.get_sentence_list(".hello.")

        self.assertEqual(["hello", "."], ret1)
        self.assertEqual(["."], ret2)
        self.assertEqual([".", "."], ret3)

    def test_trailing_whitespace(self):
        ret1 = self.tu.get_sentence_list("hello ")
        ret2 = self.tu.get_sentence_list("hello\n")
        ret3 = self.tu.get_sentence_list("hello \n")
        ret4 = self.tu.get_sentence_list("hello \n\r")
        ret5 = self.tu.get_sentence_list("hello \n\r vox ")

        self.assertEqual(["hello"], ret1)
        self.assertEqual(["hello"], ret2)
        self.assertEqual(["hello"], ret3)
        self.assertEqual(["hello"], ret4)
        self.assertEqual(["hello", "vox"], ret5)


class test_filename_from_sent(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        th.create_test_files(test_filepath, test_files)

    @classmethod
    def tearDownClass(cls):
        th.clean_test_files(test_filepath, test_exportpath)

    def setUp(self):
        self.tu = Voice(test_files_paths["normal"],
                        test_exportpath_p.joinpath("normal"))

    def tearDown(self):
        self.tu.exit()

    def test_not_gen_sent(self):
        fp = self.tu.get_filepath_from_sent("hello")

        self.assertEqual(fp, None)

    def test_simple_sent(self):
        self.tu.generate_audio("hello")

        fp = self.tu.get_filepath_from_sent("hello")

        self.assertEqual(fp, test_exportpath_p.joinpath(
            "normal").joinpath("1.wav"))


class test_sayable_unsayable(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        th.create_test_files(test_filepath, test_files)

    @classmethod
    def tearDownClass(cls):
        th.clean_test_files(test_filepath, test_exportpath)

    def setUp(self):
        self.tu = Voice(test_files_paths["normal"],
                        test_exportpath_p.joinpath("normal"))

    def tearDown(self):
        self.tu.exit()

    def test_empty_sent(self):
        ret_say = self.tu.get_sayable("")
        ret_unsay = self.tu.get_unsayable("")

        self.assertEqual([], ret_say)
        self.assertEqual([], ret_unsay)

    def test_simple_sent(self):
        ret_say = self.tu.get_sayable("hello")
        ret_unsay = self.tu.get_unsayable("hello")

        self.assertEqual(["hello"], ret_say)
        self.assertEqual([], ret_unsay)

    def test_duplicates(self):
        sent = "hello hello world world , , . . duplicates! duplicates"

        ret_say = self.tu.get_sayable(sent)
        ret_unsay = self.tu.get_unsayable(sent)

        self.assertCountEqual(["hello", ",", "."], ret_say)
        self.assertCountEqual(
            ["world", "duplicates", "duplicates!"], ret_unsay)

    def test_comp_sent(self):
        sent = "hello, world. Vox can't say some of this."

        ret_say = self.tu.get_sayable(sent)
        ret_unsay = self.tu.get_unsayable(sent)

        self.assertCountEqual(["hello", ",", "vox", "."], ret_say)
        self.assertCountEqual(
            ["world", "can't", "say", "some", "of", "this"], ret_unsay)

    def test_dup_punct(self):
        sent = "hello... world"

        ret_say = self.tu.get_sayable(sent)
        ret_unsay = self.tu.get_unsayable(sent)

        self.assertCountEqual(["hello", "."], ret_say)
        self.assertCountEqual(["world"], ret_unsay)


class test_sentence_generation(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        th.create_test_files(test_filepath, test_files)

    @classmethod
    def tearDownClass(cls):
        th.clean_test_files(test_filepath, test_exportpath)

    def setUp(self):
        self.tu = Voice(test_files_paths["normal"],
                        test_exportpath_p.joinpath("normal"))

        self.exp_ret = {"sentence": None,
                        "sayable": None,
                        "unsayable": None,
                        "path": None}

        self.exports = None
        self.exp_exports = None

    def tearDown(self):
        self.tu.exit()
        th.clean_test_files(test_filepath, test_exportpath,
                            db_exports_only=True)

    def test_empty_sent(self):
        ret = self.tu.generate_audio("")
        self.exports = list(test_exportpath_p.joinpath("normal").iterdir())

        self.exp_ret["sentence"] = ""
        self.exp_ret["sayable"] = []
        self.exp_ret["unsayable"] = []

        self.exp_exports = []

        self.assertEqual(self.exp_ret, ret)
        self.assertEqual(self.exp_exports, self.exports)

    def test_unsayable_sent(self):
        ret = self.tu.generate_audio("whatthefuckdidyoujustsaytome")
        self.exports = list(test_exportpath_p.joinpath("normal").iterdir())

        self.exp_ret["sentence"] = ""
        self.exp_ret["sayable"] = []
        self.exp_ret["unsayable"] = ["whatthefuckdidyoujustsaytome"]

        self.exp_exports = []

        self.assertEqual(self.exp_ret, ret)
        self.assertEqual(self.exp_exports, self.exports)

    def test_sayable_sent(self):
        sentence = "hello, my name is vox"
        ret = self.tu.generate_audio(sentence)

        # exp_path = test_exportpath_p.joinpath(
        #     "0.wav")

        exp_path = self.tu.get_filepath_from_sent("hello , my name is vox")
        self.exports = list(test_exportpath_p.joinpath("normal").iterdir())

        self.exp_ret["sentence"] = "hello , my name is vox"
        self.exp_ret["sayable"] = [",", "hello", "is", "my", "name", "vox"]
        self.exp_ret["unsayable"] = []
        self.exp_ret["path"] = exp_path

        self.exp_exports = [exp_path]

        self.assertEqual(self.exp_ret, ret)
        self.assertEqual(self.exp_exports, self.exports)

    def test_duplicate_sent(self):
        exp_path = test_exportpath_p.joinpath(
            "0.wav")
        self.tu.generate_audio("hello")
        self.tu.generate_audio("hello")

        exp_path = self.tu.get_filepath_from_sent("hello")

        self.exports = list(test_exportpath_p.joinpath("normal").iterdir())

        self.assertEqual(self.exports, [exp_path])

    def test_duplicate_words(self):
        ret = self.tu.generate_audio("hello hello hello")
        exp_path = self.tu.get_filepath_from_sent("hello hello hello")
        self.exports = list(test_exportpath_p.joinpath("normal").iterdir())

        self.exp_ret["sentence"] = "hello hello hello"
        self.exp_ret["sayable"] = ["hello"]
        self.exp_ret["unsayable"] = []
        self.exp_ret["path"] = exp_path

        self.exp_exports = [exp_path]

        self.assertEqual(self.exp_ret, ret)
        self.assertEqual(self.exp_exports, self.exports)

    def test_dup_punct(self):
        ret = self.tu.generate_audio("hello... hello")
        exp_path = self.tu.get_filepath_from_sent("hello . . . hello")
        self.exports = list(test_exportpath_p.joinpath("normal").iterdir())

        self.exp_ret["sentence"] = "hello . . . hello"
        self.exp_ret["sayable"] = [".", "hello"]
        self.exp_ret["unsayable"] = []
        self.exp_ret["path"] = exp_path

        self.exp_exports = [exp_path]

        self.assertEqual(self.exp_ret, ret)
        self.assertEqual(self.exp_exports, self.exports)

    def test_multiple_sent(self):

        self.tu.generate_audio("hello")
        self.tu.generate_audio("vox")
        self.tu.generate_audio(".")
        self.tu.generate_audio(",")

        test_paths = []
        test_paths.append(self.tu.get_filepath_from_sent("hello"))
        test_paths.append(self.tu.get_filepath_from_sent("vox"))
        test_paths.append(self.tu.get_filepath_from_sent("."))
        test_paths.append(self.tu.get_filepath_from_sent(","))

        # TODO: fix the stupid double joinpaths
        exp_paths = [test_exportpath_p.joinpath("normal").joinpath("1.wav"), test_exportpath_p.joinpath("normal").joinpath(
            "2.wav"), test_exportpath_p.joinpath("normal").joinpath("3.wav"), test_exportpath_p.joinpath("normal").joinpath("4.wav")]

        self.assertEqual(exp_paths, test_paths)

    # def test_potential_overwrite1(self):

    #     self.tu.generate_audio("hello.")
    #     self.tu.generate_audio("hello..")

    #     test_paths = []
    #     test_paths.append(self.tu.get_filepath_from_sent("hello"))
    #     test_paths.append(self.tu.get_filepath_from_sent("hello."))
    #     test_paths.append(self.tu.get_filepath_from_sent("hello.."))

    #     exp_paths = [test_exportpath_p.joinpath("normal").joinpath("1.wav"), test_exportpath_p.joinpath(
    #         "normal").joinpath("2.wav"), test_exportpath_p.joinpath("normal").joinpath("3.wav")]

    #     self.assertEqual(exp_paths, test_paths)

    # def test_potential_overwrite2(self):
    #     self.tu.generate_audio("hello")

    #     test_paths = []
    #     test_paths.append(self.tu.get_filepath_from_sent("hello"))

    #     exp_paths = [test_exportpath_p.joinpath("normal").joinpath("1.wav")]

    #     self.assertEqual(exp_paths, test_paths)

    #     self.tu.generate_audio("hello.")


class test_get_sayable_string(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        th.create_test_files(test_filepath, test_files)

    @classmethod
    def tearDownClass(cls):
        th.clean_test_files(test_filepath, test_exportpath)

    def setUp(self):
        self.tu = Voice(test_files_paths["normal"],
                        test_exportpath_p.joinpath("normal"))

    def tearDown(self):
        self.tu.exit()

    def test_simple_sent(self):
        ret = self.tu.get_sentence_string("hello")

        self.assertEqual(ret, "hello")

    def test_comp_sent(self):
        ret = self.tu.get_sentence_string("hello. my name, is, vox")

        self.assertEqual(ret, "hello . my name , is , vox")


class test_get_generated_sentences(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        th.create_test_files(test_filepath, test_files)

    @classmethod
    def tearDownClass(cls):
        th.clean_test_files(test_filepath, test_exportpath)

    def setUp(self):
        self.tu = Voice(test_files_paths["normal"],
                        test_exportpath_p.joinpath("normal"))

    def tearDown(self):
        self.tu.exit()
        th.clean_test_files(test_filepath, test_exportpath,
                            db_exports_only=True)

    def test_no_sentences(self):
        ret = self.tu.get_generated_sentences()

        self.assertEqual(ret, [])

    def test_single_sentences(self):
        self.tu.generate_audio("hello")

        ret = self.tu.get_generated_sentences()

        self.assertEqual(ret, ["hello"])

    def test_multiple_sentences(self):
        self.tu.generate_audio("hello")
        self.tu.generate_audio("vox")

        ret = self.tu.get_generated_sentences()

        self.assertEqual(ret, ["hello", "vox"])



from pathlib import Path
from pydub import AudioSegment
from tinydb import TinyDB, where
import json
import logging


class Voice:

    def __init__(self, path, export_path):
        self.logger = logging.getLogger(__name__)
        self.logger.setLevel("INFO")

        self.path = Path(path)
        self.export_path = Path(export_path)

        self.info_path = self.path.joinpath("info/")
        self.info_name = "info.json"

        self.db_path = self.path.joinpath("db/")
        self.db_name = "db.json"

        self.name = self.path.name

        self._punctuation = {"_comma": ",", "_period": "."}
        self._rev_punctuation = {",": "_comma", ".": "_period"}
        self._punctuation_timing = {",": 250, ".": 500}

        self.alarms = []

        self._db = self._setup_db(self.db_path, self.db_name)

        self._word_dict = self._build_word_dict(self.path)
        self._audio_format = self._find_audio_format(
            self._word_dict)  # TODO: Use properies?

        self._read_info(self.info_path, self.info_name)

        # TODO: Should probably go somewhere else
        self.export_path.mkdir(parents=True, exist_ok=True)
        # self._setup()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self._db.close()

    # TODO: Need a better way to handle this
    def exit(self):
        self._db.close()

    def _setup_db(self, path, db_name):
        path.mkdir(parents=True, exist_ok=True)
        return TinyDB(path.joinpath(db_name))

    def _build_word_dict(self, path) -> dict:
        word_dict = {}
        word_files = list(x for x in path.iterdir() if x.is_file())

        for word in word_files:
            word = Path(word)
            if word.stem in self._punctuation:
                name = self._punctuation[word.stem]
            else:
                name = word.stem
            word_dict[name] = word

        self._validate_dictionary(word_dict)

        return word_dict

    def _read_info(self, path, info_name):
        info_path = path.joinpath(info_name)
        if info_path.exists():
            with open(info_path, 'r') as info_file:
                info = json.load(info_file)
                self.alarms = info["alarms"]
        self.logger.info(f"Available alarms are {self.alarms}")

    def _validate_dictionary(self, word_dict):
        # Check if there are any words
        if len(word_dict) == 0:
            self.logger.error("No words found")
            raise Exception("No words found")

        for punctuation in self._punctuation.values():
            if punctuation not in word_dict:
                # TODO: This probably doesn't need to be an exception
                # it can operate fine without punctuation
                self.logger.warning("Incomplete punctuation")
                # raise Exception("Incomplete Punctuation")

    def _find_audio_format(self, word_dict) -> str:
        file_format = None
        for path in word_dict.values():
            if file_format == None:

                file_format = path.suffix[1:]
            else:
                if str(file_format) != str(path.suffix[1:]):
                    self.logger.error(
                        "Inconsistent audio formats in the word dict")
                    raise Exception("Inconsistent Audio Formats")
        self.logger.info(f"Audio format found: {file_format}")
        return file_format

    def _process_sentence(self, sentence) -> list:
        """
        Takes a normally formatted sentence and breaks it into base elements

        Args:
            sentence (string): normally formatted sentence to be processed

        Returns:
            (array): array of elements in sentence
        """
        sentence = sentence.lower()
        sentence = sentence.rstrip()
        self.logger.info(f"Processing sentence '{sentence}'")

        # Initial split on spaces
        split_sent = sentence.split(" ")

        # Pull out punctuation
        reduced_sent = []
        for item in split_sent:
            # find first punctuation mark, if any
            first_punct = next(
                (punct for punct in self._punctuation.values() if punct in item), False)

            if first_punct:
                # Get its index
                first_punct_ind = item.find(first_punct)

                # Add everything before punct (the word)
                reduced_sent.append(item[:first_punct_ind])

                # Add all the punctuation if its actually punctuation
                # TODO: Figure out if I want to deal with types like ".hello" throwing out all the characters after the period.
                [reduced_sent.append(punct)
                 for punct in item[first_punct_ind:] if punct in self._rev_punctuation.keys()]

            else:
                reduced_sent.append(item)

        # Clean blanks from reduced_sent
        if '' in reduced_sent:
            reduced_sent = [value for value in reduced_sent if value != '']

        self.logger.info(f"Sentence processed: '{reduced_sent}'")
        return reduced_sent

    def _get_sayable_words(self, processed_sentence_arr, word_dict) -> list:
        sayable = [word for word in processed_sentence_arr if word in word_dict]
        sayable_dedup = list(dict.fromkeys(sayable))  # removes duplicates
        sayable_dedup.sort()
        return sayable_dedup

    def _get_unsayable_words(self, processed_sentence_arr, sayable_words) -> list:
        unsayable = list(set(processed_sentence_arr) - set(sayable_words))
        return list(dict.fromkeys(unsayable))  # removes duplicates

    def _get_sayable_sentence_arr(self, processed_sentence_arr, sayable_words) -> list:
        return [word for word in processed_sentence_arr if word in sayable_words]

    def _create_sentence_string(self, words) -> str:
        if len(words) == 1:
            return words[0]
        else:
            return " ".join(words)

    def get_words(self) -> list:
        """Gets the available words for the voice
        Returns:
            (list): Words available to the voice
        """
        word_list = list(self._word_dict.keys())
        word_list.sort()
        return word_list

    def get_audio_format(self) -> str:
        """Get the audio format of the voice files as well as generated files
        Returns:
            (string): Audio format
        """
        return self._audio_format

    def get_filepath_from_sent(self, sentence):
        """
        Gets the filepath of an already generated sentence

        Args:
            sentence (string): The sentence to search for. Should be normally formatted already.
        Returns:
            (pathlib): Path to the desired sentence, or None if not found.
        """
        sentence = self.get_sentence_string(sentence)
        q = self._db.search(where("sentence") == sentence)

        if len(q) == 0:
            return None
        elif len(q) == 1:
            return Path(self.export_path, str(q[0].doc_id) + "." + self._audio_format)
        else:
            self.logger.error("Multiple entries for same sentence in DB")
            raise Exception("DB Error")

    def get_sayable(self, sentence) -> list:
        """Gets the words that the voice can say from a sentence

        Args:
            sentence (string): Sentence to check for sayable words
        Returns:
            (list): Words that can be said
        """

        proc_sentence = self._process_sentence(sentence)
        return self._get_sayable_words(proc_sentence, self._word_dict)

    def get_unsayable(self, sentence) -> list:
        """Gets the words that the voice cannot say from a sentence

        Args:
            sentence (string): Sentence to check for un-sayable words
        Returns:
            (list): Words that cannot be said
        """

        proc_sentence = self._process_sentence(sentence)
        sayable_words = self._get_sayable_words(proc_sentence, self._word_dict)
        return self._get_unsayable_words(proc_sentence, sayable_words)

    def get_sentence_list(self, sentence) -> list:
        """Converts sentence into list of words that can be said

        Args:
            sentence (string): Sentence to convert to list
        Returns:
            (list): Sentence in list format excluding words that cannot be said
        """

        proc_sentence = self._process_sentence(sentence)
        sayable_words = self._get_sayable_words(proc_sentence, self._word_dict)
        return self._get_sayable_sentence_arr(proc_sentence, sayable_words)

    def get_sentence_string(self, sentence):
        proc_sentence = self._process_sentence(sentence)
        sayable_words = self._get_sayable_words(
            proc_sentence, self._word_dict)
        sayable_sent_arr = self._get_sayable_sentence_arr(
            proc_sentence, sayable_words)

        return self._create_sentence_string(sayable_sent_arr)

    def get_generated_sentences(self) -> list:
        """Gets the previously generated sentence strings

        Returns:
            (list): List of sentence strings generated previously
        """
        return [entry["sentence"] for entry in self._db.all()]

    def generate_audio(self, sentence) -> dict:
        """Generates an audio file from the given sentence

        Args:
            sentence (string): Sentence string to be generated
            start_pad (int): Silence in ms to pad beginning of file with
        Returns:
            (dict): Information about generated sentence with the following format:
            {"sentence": (string) The converted, normally formatted sentence string that was generated,
             "sayable": (list) The words in the sentence that are sayable,
             "unsayable": (list) The words in the sentence that are unsayable,
             "path": (pathlib) The path to the generated file, or None if no sentence can be created}
        """

        self.logger.info(f"Asked to generate {sentence}")
        proc_sentence = self._process_sentence(sentence)

        sayable_words = self._get_sayable_words(proc_sentence, self._word_dict)
        sayable_sent_arr = self._get_sayable_sentence_arr(
            proc_sentence, sayable_words)
        sayable_sent_str = self._create_sentence_string(sayable_sent_arr)
        unsayable_worlds = self._get_unsayable_words(
            proc_sentence, sayable_words)

        # filename = self._create_filename(sayable_sent_arr)
        # sentence_info = self._process_sentence(sentence)

        self.logger.debug(f"Generating {sayable_sent_str}")

        words_audio = []
        save_path = None

        # Only create sentence if there are words to put in it
        if len(sayable_words) != 0:

            # Search in DB for sentence string
            q = self._db.search(where("sentence") == sayable_sent_str)
            # if it doesn't exist, add it and get the ID to use as filename, then create file and save
            # if it does exist, return the ID as a path to the file
            self.logger.debug(f"Looking for {sayable_sent_str} Found: {q}")

            if len(q) == 1:
                self.logger.info(
                    f"Sentence {sayable_sent_str} already exists, not recreating")
                save_path = Path(self.export_path,
                                 str(q[0].doc_id) + "." + self._audio_format)

            elif len(q) == 0:
                sent_id = self._db.insert({"sentence": sayable_sent_str})

                for word in sayable_sent_arr:
                    if word in self._punctuation_timing:
                        words_audio.append(AudioSegment.silent(
                            self._punctuation_timing[word]))
                    else:
                        words_audio.append(AudioSegment.from_file(
                            self._word_dict[word], self._audio_format))

                sentence_audio = words_audio.pop(0)
                for word_audio in words_audio:
                    sentence_audio = sentence_audio + word_audio

                save_path = Path(self.export_path, str(
                    sent_id) + "." + self._audio_format)
                sentence_audio.export(save_path, format=self._audio_format)

            else:
                self.logger.error(
                    f"Database error. Num of queries for sentence is {len(q)}")

        else:
            self.logger.warning(
                f"Can't say any words in {sentence}, not generating")

        # If sentence isn't created, path will be None type.
        # TODO: None path might need rethinking. Maybe a generated flag?
        return {"sentence": sayable_sent_str,
                "sayable": sayable_words,
                "unsayable": unsayable_worlds,
                "path": save_path}
